#ifndef TYPON_EVENTCOUNT_HPP_INCLUDED
#define TYPON_EVENTCOUNT_HPP_INCLUDED

#include <atomic>
#include <cstdint>

namespace typon
{

  template <unsigned int N = 10>
  struct EventCount
  {
    using u64 = std::uint_fast64_t;

    static_assert(N < 16);

    static constexpr u64 shift = u64(1) << N;
    static constexpr u64 mask = shift - 1;

    std::atomic<u64> _state {0};

    auto prepare_wait() noexcept
    {
      return _state.fetch_add(1, std::memory_order_acq_rel);
    }

    void cancel_wait() noexcept
    {
      _state.fetch_sub(1, std::memory_order_seq_cst);
    }

    void wait(u64 state) noexcept
    {
      for(;;)
      {
        _state.wait(state, std::memory_order_acquire);
        auto newstate = _state.load(std::memory_order_acquire);
        if ((newstate ^ state) & (~mask))
        {
          break;
        }
        state = newstate;
      }
      _state.fetch_sub(1, std::memory_order_seq_cst);
    }

    void notify_one() noexcept
    {
      _state.fetch_add(shift, std::memory_order_acq_rel);
      _state.notify_one();
    }

    void notify_all() noexcept
    {
      _state.fetch_add(shift, std::memory_order_acq_rel);
      _state.notify_all();
    }
  };

}

#endif // TYPON_EVENTCOUNT_HPP_INCLUDED
